import Note from '../../models/Note'
import Folder from '../../models/Folder'
import {
	Paginated,
	PaginationOptions,
	StoragePluginBase,
	StoragePlugin,
	StoragePluginOptions,
} from '../../modules/Plugin/service'

interface LocalStoragePluginOptions extends StoragePluginOptions {
	noteStorageId?: string,
	folderStorageId?: string,
	engine?: Storage,
}

class LocalStoragePlugin extends StoragePluginBase implements StoragePlugin {
	private readonly engine: Storage
	private readonly noteStorageId: string
	private readonly folderStorageId: string
	private noteCache: Note[]
	private folderCache: Folder[]
	private readonly serialize = JSON.stringify
	private readonly deserialize = JSON.parse

	constructor({
		engine = localStorage,
		noteStorageId = 'zeichen.notes',
		folderStorageId = 'zeichen.folders',
	}: LocalStoragePluginOptions) {
		super()
		this.engine = engine
		this.noteStorageId = noteStorageId
		this.folderStorageId = folderStorageId
	}

	private async syncFolders() {
		if (!this.folderCache) {
			const foldersSerialized: string = await this.engine.getItem(this.folderStorageId) || '[]'
			const folders: Folder[] = this.deserialize(foldersSerialized)
			folders.forEach((folder, _, thisCollection) => {
				if (folder.parent) {
					for (let i = 0; i < thisCollection.length; i += 1) {
						if (thisCollection[i].id !== folder.parent.id) {
							continue
						}
						folder.parent = thisCollection[i]
					}
				}
			})
			this.folderCache = folders
		}
	}

	private async syncNotes() {
		if (!this.noteCache) {
			await this.syncFolders()
			const notesSerialized: string = await this.engine.getItem(this.noteStorageId) || '[]'
			const notesRaw: Note[] = this.deserialize(notesSerialized)
			this.noteCache = notesRaw.map(note => ({
				...note,
				folder: this.folderCache.find(f => f.id === note.folder.id)
			}))
		}
	}

	async saveNote(newNote: Note) {
		await this.syncNotes()
		const replacementNote = {
			...newNote,
			folder: this.folderCache.find(f => f.id === newNote.folder.id)
		}
		const targetNote = this.noteCache.some(oldNote => oldNote.id === newNote.id)
		const newNotes = (
			targetNote
				? this.noteCache.map(oldNote => (
					oldNote.id === newNote.id
						? replacementNote
						: oldNote
				))
				: [...this.noteCache, replacementNote]
		)
		await this.engine.setItem(this.noteStorageId, this.serialize(newNotes))
		this.noteCache = newNotes
		return replacementNote
	}

	async loadNote(noteId: string) {
		await this.syncNotes()
		return this.noteCache.find(cachedNote => cachedNote.id === noteId)
	}

	async loadAllNotes(options = {} as PaginationOptions) {
		await this.syncNotes()
		const { skip, take } = options
		const data = {
			...this.noteCache.slice(skip, skip + take),
			totalLength: this.noteCache.length,
		}
		return data as Paginated<Note>
	}

	async loadFolder(folderId: string) {
		await this.syncFolders()
		return this.folderCache.find(cachedFolder => cachedFolder.id === folderId)
	}

	async saveFolder(newFolder: Folder) {
		await this.syncFolders()
		const replacementFolder = {
			...newFolder,
			parent: this.folderCache.find(f => f.id === newFolder.parent.id)
		}
		const targetFolder = this.folderCache.some(oldFolder => oldFolder.id === newFolder.id)
		const newFolders = (
			targetFolder
				? this.folderCache.map(oldFolder => (
					oldFolder.id === newFolder.id
						? replacementFolder
						: oldFolder
				))
				: [...this.folderCache, replacementFolder]
		)
		const newFoldersSerializeObject = newFolders.map(f => ({
			...f,
			parent: { id: f.parent.id, },
		}))
		await this.engine.setItem(this.folderStorageId, this.serialize(newFoldersSerializeObject))
		this.folderCache = newFolders
		return replacementFolder
	}
}

interface RemoteStoragePluginOptions extends StoragePluginOptions {
	baseUrl: string,
}

class RemoteStoragePlugin extends StoragePluginBase implements StoragePlugin {
	private readonly baseUrl: string
	private noteCache: Note[]
	private folderCache: Folder[]

	constructor({
		baseUrl,
	}: RemoteStoragePluginOptions) {
		super()
		this.baseUrl = baseUrl
	}

	private async syncFolders() {
		if (!this.folderCache) {
			const url = new URL('/folders', this.baseUrl)
			const response = await fetch(url.toString())
			const folders: Folder[] = await response.json()
			folders.forEach((folder, _, thisCollection) => {
				if (folder.parent) {
					for (let i = 0; i < thisCollection.length; i += 1) {
						if (thisCollection[i].id !== folder.parent.id) {
							continue
						}
						folder.parent = thisCollection[i]
					}
				}
			})
			this.folderCache = folders
		}
	}

	private async syncNotes() {
		if (!this.noteCache) {
			await this.syncFolders()
			const url = new URL('/notes', this.baseUrl)
			const response = await fetch(url.toString())
			const notesRaw: Note[] = await response.json()
			this.noteCache = notesRaw.map(note => ({
				...note,
				folder: this.folderCache.find(f => f.id === note.folder.id)
			}))
		}
	}

	async loadNote(noteId: string) {
		await this.syncNotes()
		return this.noteCache.find(cachedNote => cachedNote.id === noteId)
	}

	async loadAllNotes(options = {} as PaginationOptions) {
		await this.syncNotes()
		const { skip, take } = options
		const data = {
			...this.noteCache.slice(skip, skip + take),
			totalLength: this.noteCache.length,
		}
		return data as Paginated<Note>
	}

	async loadFolder(folderId: string) {
		await this.syncFolders()
		return this.folderCache.find(cachedFolder => cachedFolder.id === folderId)
	}

	async saveNote(newNote: Note) {
		await this.syncNotes()
		const relativeUrl: string = Boolean(newNote.id) ? `/notes/${newNote.id}` : '/notes'
		const method = Boolean(newNote.id) ? 'PUT' : 'POST'
		const url = new URL(relativeUrl, this.baseUrl)
		const response = await fetch(url.toString(), {
			method,
			body: JSON.stringify(newNote),
		})
		const retrieved: Note = await response.json()
		const replacementNote: Note = {
			...retrieved,
			folder: this.folderCache.find(f => f.id === newNote.folder.id)
		}
		const targetNote = this.noteCache.some(oldNote => oldNote.id === newNote.id)
		this.noteCache = (
			targetNote
				? this.noteCache.map(oldNote => (
					oldNote.id === newNote.id
						? replacementNote
						: oldNote
				))
				: [...this.noteCache, replacementNote]
		)
		return replacementNote
	}

	async saveFolder(newFolder: Folder) {
		await this.syncFolders()
		const relativeUrl: string = Boolean(newFolder.id) ? `/folders/${newFolder.id}` : '/notes'
		const method = Boolean(newFolder.id) ? 'PUT' : 'POST'
		const url = new URL(relativeUrl, this.baseUrl)
		const response = await fetch(url.toString(), {
			method,
			body: JSON.stringify(newFolder),
		})
		const retrieved: Folder = await response.json()
		const replacementFolder: Folder = {
			...retrieved,
			parent: this.folderCache.find(f => f.id === newFolder.parent.id)
		}
		const targetFolder = this.folderCache.some(oldFolder => oldFolder.id === newFolder.id)
		this.folderCache = (
			targetFolder
				? this.folderCache.map(oldFolder => (
					oldFolder.id === newFolder.id
						? replacementFolder
						: oldFolder
				))
				: [...this.folderCache, replacementFolder]
		)
		return replacementFolder
	}
}
